﻿CompilerIf Defined(INCLUDE_ARRAYCLASS, #PB_Constant)=0
#INCLUDE_ARRAYCLASS=1
;/////////////////////////////////////////////////////////////////////////////////
;An OOP class for implementing either a simple dynamic array or a structured array.

;By Stephen Rodriguez.
;January 2010.

;Developed with Purebasic 4.4.
;Fully cross-platform.
;/////////////////////////////////////////////////////////////////////////////////
;/////////////////////////////////////////////////////////////////////////////////
;To create an instance of a new array class, use something like :
;
;   MyArrayObject.ArrayObject
;   MyArrayObject = NewArray(numberOfElements)
;
;and check the return value is non zero.
;
;For a structured array use :
;
;   MyArrayObject.StructuredArrayObject
;   MyArrayObject = NewStructuredArray(numElements, sizeOfEachElement, numberOfStringFields)
;
;and check the return value is non zero.
;
;NOTE: for a structured array, ALL string fields must be placed before all other fields.
;/////////////////////////////////////////////////////////////////////////////////
;/////////////////////////////////////////////////////////////////////////////////
;Methods exposed by both of the array classes :
;
;      Destroy()
;      GetArrayBase()
;      GetUpperBound()
;      GetValue.i(index)
;      ReDim.i(numElements)
;      SetValue.i(index, value)
;      ShiftElementsLeft(index)
;      ShiftElementsRight(index)
;      SwapElements(index1, index2)
;
;Additional method exposed by the structured array class :
;
;      GetPointerToValue.i(index)
;/////////////////////////////////////////////////////////////////////////////////


XIncludeFile "..\array\arrayClass_Residents.pbi"

;The following structure contains the class template and private properties,
  Structure _membersArrayClass
    *vTable.i
    UpperBound.i
    ArrayBase.i
    ;The following field(s) are used by structured array objects.
      SizeOfEachElement.i
      NumberOfStringFields.i  ;All string fields must reside at the beginning of the
                              ;structure.
  EndStructure 


Declare.i StructuredArrayClass_SetValue(*this._membersArrayClass, index, ptrStructure=0)


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of the basic arrayClass with elements indexed from 0 to numElements.
;Returns zero if an error.
Procedure.i NewArray(uBound)
  Protected *object._membersArrayClass, base
  If uBound >= 0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersArrayClass))
      If *object
        ;Allocate memory for the array.
          base = AllocateMemory((uBound+1)*SizeOf(INTEGER))
          If base
            ;Set the object properties.
              *object\vTable = ?VTable_ArrayClass
              *object\UpperBound = uBound
              *object\ArrayBase = base
          Else
            FreeMemory(*object)
            *object = 0          
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of a structured arrayClass.
;Returns zero if an error.
Procedure.i NewStructuredArray(uBound, SizeOfEachElement, NumberOfStringFields)
  Protected *object._membersArrayClass, base, i, *ptrInteger.INTEGER, blnError
  If uBound>=0 And SizeOfEachElement And NumberOfStringFields>=0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersArrayClass))
      If *object
        ;Allocate memory for the array.
          base = AllocateMemory((uBound+1)*SizeOf(INTEGER))
          If base
            ;Set the object properties.
              *object\vTable = ?VTable_StructuredArrayClass
              *object\UpperBound = uBound
              *object\ArrayBase = base
              *object\SizeOfEachElement = SizeOfEachElement
              *object\NumberOfStringFields = NumberOfStringFields
              ;Now allocate the memory for each array element (ignoring the string fields).
                *ptrInteger = base
                For i = 0 To uBound
                  *ptrInteger\i = AllocateMemory(SizeOfEachElement)
                  If *ptrInteger\i = 0 ;Error.
                    blnError = #True
                    Break
                  EndIf
                  *ptrInteger + SizeOf(INTEGER)                
                Next
              If blnError
                ;Free all memory and report an error.
                  While i
                    *ptrInteger-SizeOf(INTEGER)
                    FreeMemory(*ptrInteger\i)
                    i-1
                  Wend              
                  FreeMemory(base)
                  FreeMemory(*object)
                  *object = 0
              EndIf
          Else
            FreeMemory(*object)
            *object = 0          
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



;-BASE CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following returns the base address of the memory buffer containing the array elements of the underlying array.
Procedure.i arrayBaseClass_GetArrayBase(*this._membersArrayClass)
  ProcedureReturn *this\ArrayBase
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the upper bound of the underlying array.
Procedure.i arrayBaseClass_GetUpperBound(*this._membersArrayClass)
  ProcedureReturn *this\UpperBound
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method swaps two elements in the underlying array.
Procedure arrayBaseClass_SwapElements(*this._membersArrayClass, index1, index2)
  Protected *ptr1Integer.INTEGER, *ptr2Integer.INTEGER, temp
  If index1 >= 0 And index2 >=0 And index1 <= *this\UpperBound And index2 <= *this\UpperBound And index1<>index2
    *ptr1Integer = *this\ArrayBase + index1*SizeOf(INTEGER)
    *ptr2Integer = *this\ArrayBase + index2*SizeOf(INTEGER)
    temp = *ptr1Integer\i
    *ptr1Integer\i = *ptr2Integer\i
    *ptr2Integer\i = temp
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-BASIC ARRAY CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with the underlying array.
Procedure arrayClass_Destroy(*this._membersArrayClass)
  FreeMemory(*this\ArrayBase)
  FreeMemory(*this)
  *this = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method retrieves a value from the underlying array.
Procedure.i arrayClass_GetValue(*this._membersArrayClass, index)
  Protected *ptrInteger.INTEGER
  If index >= 0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    ProcedureReturn *ptrInteger\i
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following writes the given value at the specified position in the underlying array.
;Returns zero if an error, i.e. if the index is out of bounds etc.
Procedure.i arrayClass_SetValue(*this._membersArrayClass, index, Value)
  Protected *ptrInteger.INTEGER
  If index >= 0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    *ptrInteger\i = Value
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method shifts all elements of the underlying array, starting from the last element
;and up to the specified element along towards the beginning (top) of the array. This effectively deletes an element.
;The newly vacated 'last' element is given the value of zero.
;The array dimension is not altered etc.
Procedure arrayClass_ShiftElementsLeft(*this._membersArrayClass, index)
  Protected *ptrInteger.INTEGER, temp, i
  If index>=0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    *ptrInteger\i = 0 ;In case only the last element is being shifted.
    For i = index To *this\UpperBound-1
      *ptrInteger\i = PeekI(*ptrInteger+SizeOf(INTEGER))
      *ptrInteger+SizeOf(INTEGER)
    Next
    *ptrInteger\i = 0
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method shifts all elements of the underlying array, starting from the specified
;index, along towards the end (bottom) of the array. This effectively inserts an element.
;The final element is discarded and the newly vacated first element is given the value of zero.
;The array dimension is not altered etc.
Procedure arrayClass_ShiftElementsRight(*this._membersArrayClass, index)
  Protected *ptrInteger.INTEGER, temp, i
  If index>=0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + *this\UpperBound*SizeOf(INTEGER)
    *ptrInteger\i = 0 ;In case only the last element is being shifted.
    For i = *this\UpperBound-1 To index Step-1
      *ptrInteger\i = PeekI(*ptrInteger-SizeOf(INTEGER))
      *ptrInteger-SizeOf(INTEGER)
    Next
    *ptrInteger\i = 0
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method redimensions the underlying array whilst preserving the array contents.
;In the case of an error, zero is returned but the original array remains valid.
;Returns non zero if there is no error.
Procedure.i arrayClass_ReDim(*this._membersArrayClass, uBound)
  Protected mem
  If uBound >= 0 
    If *this\UpperBound <> uBound
      mem = ReAllocateMemory(*this\ArrayBase, (uBound+1)*SizeOf(INTEGER))
      If mem ;Reallocation was successful.
        *this\UpperBound = uBound
        *this\ArrayBase = mem
        ProcedureReturn #True
      Else
        ProcedureReturn #False
      EndIf
    Else
      ProcedureReturn #True
    EndIf    
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-STRUCTURED ARRAY CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with the underlying array.
Procedure StructuredarrayClass_Destroy(*this._membersArrayClass)
  Protected i, *ptrInteger.INTEGER
  ;Free the memory allocated for each element.
    *ptrInteger = *this\ArrayBase
    For i = 0 To *this\UpperBound
      ;Free the strings for this element.
        StructuredArrayClass_SetValue(*this, i, 0) 
      ;Now free the memory for the element itself.
        FreeMemory(*ptrInteger\i)
      *ptrInteger+SizeOf(INTEGER)
    Next
  FreeMemory(*this\ArrayBase)
  FreeMemory(*this)
  *this = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method retrieves a structure from the underlying array.
Procedure StructuredArrayClass_GetValue(*this._membersArrayClass, index, ptrStructure)
  Protected *ptrInteger.INTEGER, *ptrStr.String, memStruct, i, text$
  If index>=0 And index <=*this\UpperBound And ptrStructure
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    memStruct = *ptrInteger\i
    ;Sort out any string fields.
      *ptrStr = ptrStructure : *ptrInteger = memStruct
      For i = 1 To *this\NumberOfStringFields
        If *ptrInteger\i
          text$ = PeekS(*ptrInteger\i)
          ;We need to copy this string to the structure specified by ptrStructure.
          ;We get PB to do this in such a way that the heap memory is allocated by PB itself
          ;so that it knows to free it as appropriate and also frees any existing string in the
          ;structure being overwritten.
            *ptrStr\s = text$
        ElseIf PeekI(*ptrStr)
          *ptrStr\s = ""  ;This means any existing string will be freed by PB.
        EndIf
        *ptrStr + SizeOf(INTEGER)
        *ptrInteger+SizeOf(INTEGER)
      Next
    ;Copy the remaining fields; those following the string fields.
      CopyMemory(memStruct+*this\NumberOfStringFields*SizeOf(INTEGER), ptrStructure+*this\NumberOfStringFields*SizeOf(INTEGER), *this\SizeOfEachElement-*this\NumberOfStringFields*SizeOf(INTEGER))
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method retrieves a pointer to a structure stored in the underlying array.
Procedure.i StructuredArrayClass_GetPointerToValue(*this._membersArrayClass, index)
  Protected result, *ptrInteger.INTEGER
  If index>=0 And index <=*this\UpperBound
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    result = *ptrInteger\i
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following writes the given structure at the specified position in the underlying array.
;If ptrStructure = 0 then the specified element is zeroed (and all strings freed).
;Returns zero if an error, i.e. if the index is out of bounds etc. Else returns a pointer to
;the newly added structure.
Procedure.i StructuredArrayClass_SetValue(*this._membersArrayClass, index, ptrStructure=0)
  Protected *ptrInteger.INTEGER, *ptrString.String, memStruct, i, text$, blnError, *ptrByte.BYTE
  Protected tempStruct
  If index>=0 And index <=*this\UpperBound
    If ptrStructure = 0
      ;The user simply wishes to zero / free the specified element.
      ;First free up the strings in the existing element.
        *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
        memStruct = *ptrInteger\i  ;The address of the structure which was written to the array.
        *ptrInteger = memStruct
        For i = 1 To *this\NumberOfStringFields
          If *ptrInteger\i
            FreeMemory(*ptrInteger\i)
          EndIf
          *ptrInteger+SizeOf(INTEGER)
        Next
        ;Got to keep this cross-platform, so no use of ZeroMemory etc.
          *ptrByte = memStruct
          For i = 0 To *this\SizeOfEachElement-1
            *ptrByte\b = 0
            *ptrByte + SizeOf(BYTE)
          Next
          ProcedureReturn memStruct
    Else
      ;The user wishes to overwrite the specified element with a new one.
      ;To avoid destroying the current element if we later encounter a memory allocation error,
      ;we copy the new structure to a temporary store first.
        tempStruct = AllocateMemory(*this\SizeOfEachElement)
        If tempStruct = 0  ;Memory allocation problem.
          ProcedureReturn #False
        EndIf
        CopyMemory(ptrStructure, tempStruct, *this\SizeOfEachElement)
      ;Now copy the strings.
        *ptrString = ptrStructure : *ptrInteger = tempStruct
        For i = 1 To *this\NumberOfStringFields
          If PeekI(*ptrString)
            ;Copy the string.
              text$ = *ptrString\s
              *ptrInteger\i = AllocateMemory((Len(text$)+1)*SizeOf(CHARACTER))
              If *ptrInteger\i
                PokeS(*ptrInteger\i, text$)
              Else
                blnError = #True
                Break
              EndIf
          EndIf
          *ptrString + SizeOf(INTEGER)
          *ptrInteger + SizeOf(INTEGER)
        Next
        If blnError
          ;We need to release all memory allocated for this element and return an error.
            While i>1
              i-1
              *ptrInteger - SizeOf(INTEGER)
              FreeMemory(*ptrInteger\i)
            Wend
            FreeMemory(tempStruct)
            ProcedureReturn #False
        EndIf
      ;No error with memory allocation so now we can copy the temporary structure to it's intended location.
        ;Begin by freeing the original strings -if any.
          *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
          memStruct = *ptrInteger\i  ;The address of the structure which was written to the array.
          *ptrInteger = memStruct
          For i = 1 To *this\NumberOfStringFields
            If *ptrInteger\i
              FreeMemory(*ptrInteger\i)
            EndIf
            *ptrInteger+SizeOf(INTEGER)
          Next
        CopyMemory(tempStruct, memStruct, *this\SizeOfEachElement)
        FreeMemory(tempStruct)
        ProcedureReturn memStruct
    EndIf
  Else
    ProcedureReturn #False
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method shifts all elements of the underlying array, starting from the last element
;and up to the specified element along towards the beginning (top) of the array. This effectively deletes an element.
;The newly vacated 'last' element is filled with an 'empty' structure.
;The array dimension is not altered etc.
Procedure StructuredArrayClass_ShiftElementsLeft(*this._membersArrayClass, index)
  Protected *ptrInteger.INTEGER, i, memStruct
  If index>=0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + index*SizeOf(INTEGER)
    memStruct = *ptrInteger\i
    ;Free the strings associated with this element.
    ;We keep the memory and use it for the vacated last element.
      StructuredArrayClass_SetValue(*this, index, 0)
    For i = index To *this\UpperBound-1
      *ptrInteger\i = PeekI(*ptrInteger+SizeOf(INTEGER))
      *ptrInteger+SizeOf(INTEGER)
    Next
    *ptrInteger\i = memStruct
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method shifts all elements of the underlying array, starting from the specified
;index, along towards the end (bottom) of the array. This effectively inserts an element.
;The final element is discarded and the newly vacated first element is filled with an 'empty' structure.
;The array dimension is not altered etc.
Procedure StructuredArrayClass_ShiftElementsRight(*this._membersArrayClass, index)
  Protected *ptrInteger.INTEGER, i, memStruct
  If index>=0 And index <= *this\UpperBound
    *ptrInteger = *this\ArrayBase + *this\UpperBound*SizeOf(INTEGER)
    memStruct = *ptrInteger\i
    ;Free the strings associated with the final element.
    ;We keep the memory and use it for the vacated first element.
      StructuredArrayClass_SetValue(*this, *this\UpperBound, 0)
    For i = *this\UpperBound-1 To index Step-1
      *ptrInteger\i = PeekI(*ptrInteger-SizeOf(INTEGER))
      *ptrInteger-SizeOf(INTEGER)
    Next
    *ptrInteger\i = memStruct
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following method redimensions the underlying array whilst preserving the array contents.
;In the case of an error, zero is returned but the original array remains valid.
;Returns non zero if there is no error.
Procedure.i StructuredArrayClass_ReDim(*this._membersArrayClass, uBound)
  Protected mem, i, *ptrInteger.INTEGER, blnError
  If uBound >= 0
    If *this\UpperBound = uBound
      ProcedureReturn #True
    EndIf
    ;We take care when enlarging the array because of the memory requirements.
    If uBound > *this\UpperBound
      ;Because of possible memory allocation problems later on, we create a new array of pointers
      ;(not the elements themselves).
        mem = AllocateMemory((uBound+1)*SizeOf(INTEGER))
      If mem
        ;Copy the existing set of pointers.
          CopyMemory(*this\ArrayBase, mem, (*this\UpperBound+1)*SizeOf(INTEGER))
        ;Create additional structures to fill the remainder of the array.
          *ptrInteger = mem + (*this\UpperBound+1)*SizeOf(INTEGER)
          For i = 0 To uBound - *this\UpperBound - 1
            *ptrInteger\i = AllocateMemory(*this\SizeOfEachElement)
            If *ptrInteger\i = 0
              blnError = #True
              Break
            EndIf
            *ptrInteger+SizeOf(INTEGER)
          Next
          If blnError
            While i
              *ptrInteger-SizeOf(INTEGER)
              FreeMemory(*ptrInteger\i)
              i-1
            Wend
            FreeMemory(mem)
            ProcedureReturn #False
          EndIf
        ;Replace the old array with this new and extended one and reset object properties.
          FreeMemory(*this\ArrayBase)
          *this\ArrayBase = mem
          *this\UpperBound = uBound
        ProcedureReturn #True
      Else
        ProcedureReturn #False
      EndIf
    Else ;The new array is smaller than the original.
      ;First release the memory for the elements to be discarded.
        *ptrInteger = *this\ArrayBase + uBound*SizeOf(INTEGER)
        For i = uBound+1 To *this\UpperBound
          *ptrInteger+SizeOf(INTEGER)
          StructuredArrayClass_SetValue(*this, i, 0)
          FreeMemory(*ptrInteger\i)
        Next
      ;Now reallocate the memory of the array of pointers.
        mem = ReAllocateMemory(*this\ArrayBase, (uBound+1)*SizeOf(INTEGER))
      ;Reset object properties.
        *this\UpperBound = uBound
        *this\ArrayBase = mem
      ProcedureReturn #True
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VIRTUAL TABLES.

DataSection 
  VTable_ArrayClass: 
    Data.i @arrayBaseClass_GetArrayBase()
    Data.i @arrayBaseClass_GetUpperBound() 
    Data.i @arrayBaseClass_SwapElements()
    Data.i @arrayClass_Destroy()
    Data.i @arrayClass_GetValue()
    Data.i @arrayClass_SetValue()
    Data.i @arrayClass_ShiftElementsLeft()
    Data.i @arrayClass_ShiftElementsRight()
    Data.i @arrayClass_ReDim()
  VTable_StructuredArrayClass: 
    Data.i @arrayBaseClass_GetArrayBase()
    Data.i @arrayBaseClass_GetUpperBound() 
    Data.i @arrayBaseClass_SwapElements()
    Data.i @StructuredarrayClass_Destroy()
    Data.i @StructuredArrayClass_GetValue()
    Data.i @StructuredArrayClass_GetPointerToValue()
    Data.i @StructuredArrayClass_SetValue()
    Data.i @StructuredArrayClass_ShiftElementsLeft()
    Data.i @StructuredArrayClass_ShiftElementsRight()
    Data.i @StructuredArrayClass_ReDim()
EndDataSection 
CompilerEndIf

; IDE Options = PureBasic 4.40 (Windows - x86)
; CursorPosition = 39
; FirstLine = 12
; Folding = ----